
/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Copyright (c) 1982, 1986, 1991, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)in.c	8.2 (Berkeley) 11/15/93
 */

#if defined(__FreeBSD__) && __FreeBSD__ >= 3
#include "opt_inet.h"
#include "opt_inet6.h"
#include "opt_qospsas.h"
#endif
#ifdef __NetBSD__
#include "opt_inet.h"
#endif

#include <sys/param.h>
#if !(defined(__FreeBSD__) && __FreeBSD__ >= 3)
#include <sys/ioctl.h>
#endif
#include <sys/errno.h>
#include <sys/malloc.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/sockio.h>
#include <sys/systm.h>

#if !defined(__bsdi__) && !(defined(__FreeBSD__) && __FreeBSD__ < 3)
#include <sys/proc.h>
#endif
#include <sys/time.h>
#include <sys/kernel.h>
#include <sys/syslog.h>

#include <net/if.h>
#include <net/if_types.h>
#include <net/route.h>
#include <net/if_dl.h>

#include <netinet/in.h>
#include <netinet6/in6_var.h>
#include <netinet6/qospsas.h>

/*
 * Definitions and global variables 
 */

MALLOC_DEFINE(M_SRCPTABLE, "ip6_srcptable", "SAS policy table");

#define IP6_SRCPTABLE_ADD 1
#define IP6_SRCPTABLE_DEL 2
#define IP6_SRCPTABLE_SHOW 3

/* Global var which represents the default policy table */
struct ip6_srcptable *ip6_default_srcptable;

/* 
 * Definition of sysctl calls from kernel/userland interaction
 */

#if defined(__FreeBSD__) && __FreeBSD__ >= 4
static int ip6_srcptable_sysctl(SYSCTL_HANDLER_ARGS);
#else
static int ip6_srcptable_sysctl SYSCTL_HANDLER_ARGS;
#endif
#ifdef SYSCTL_DECL
SYSCTL_DECL(_net_inet6_ip6);
#endif
SYSCTL_NODE(_net_inet6_ip6, IPV6CTL_SRCPTABLE, srcptable,
	CTLFLAG_RW, ip6_srcptable_sysctl, "srcptable");


static int
#if defined(__FreeBSD__) && __FreeBSD__ >= 4
ip6_srcptable_sysctl(SYSCTL_HANDLER_ARGS)
#else
ip6_srcptable_sysctl SYSCTL_HANDLER_ARGS
#endif
{
  int error= 0, option;
  char buf[1024];  /*FIXME: hard-coded */
  struct ip6_srcptable *d, *de;
  struct ip6_srcptable *srcptable_aux;   

  /* parse the data to know what to do next */
  if ( (req->newptr) && (! req->oldptr) )
  {
    error = SYSCTL_IN( req, buf, sizeof(struct ip6_srcptable)+sizeof(char) );
    if( error ) 
      return error;

    switch( buf[0] )
    {
    case 'A': /* hard-coded */
      option= IP6_SRCPTABLE_ADD;
      break;
    case 'D': /* hard-coded */
      option= IP6_SRCPTABLE_DEL;
      break;
    default:
      return EPERM;
    }
  }
  else
  if( (! req->newptr) && (req->oldptr) )
    option= IP6_SRCPTABLE_SHOW;
  else
    return EPERM;

   
  /* execute the solicited action */
  switch( option )
  {
  case IP6_SRCPTABLE_ADD:
    error= ip6_add_srcptable( (struct ip6_srcptable *)(buf+sizeof(char)) );  
    break;

  case IP6_SRCPTABLE_DEL:
    error= ip6_del_srcptable( (struct ip6_srcptable *)(buf+sizeof(char)) );    
    break;

  case IP6_SRCPTABLE_SHOW:
    for (srcptable_aux = ip6_default_srcptable; 
	 srcptable_aux; 
	 srcptable_aux = srcptable_aux->next) 
    {
      d = (struct ip6_srcptable *)buf;
      de = (struct ip6_srcptable *)(buf + sizeof(buf));

      if (d + 1 <= de) 
      {    
	bcopy( srcptable_aux, d, sizeof(*d) );
	d->next= 0;
      }
      else
	return ENOMEM;

      error = SYSCTL_OUT(req, buf, sizeof(*d));
      if (error)
	return error;
    } 
    break;

  default:
    return EOPNOTSUPP; 
  }

  return error;
}

#define IN6ADDR_SRCPTABLE_ENTRY_1 \
	{{{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
	    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01 }}}

#define IN6ADDR_SRCPTABLE_ENTRY_2 \
        {{{ 0x20, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }}}      

#define IN6ADDR_SRCPTABLE_ENTRY_3 \
        {{{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
            0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x00, 0x00 }}} 

void ip6_init_srcptable()
{
  struct ip6_srcptable *aux;
  const struct in6_addr entry_1 =  IN6ADDR_SRCPTABLE_ENTRY_1;  
  const struct in6_addr entry_2 =  IN6ADDR_SRCPTABLE_ENTRY_2; 
  const struct in6_addr entry_3 =  IN6ADDR_SRCPTABLE_ENTRY_3; 

 
  ip6_default_srcptable= (struct ip6_srcptable *)
    malloc( sizeof(struct ip6_srcptable), M_SRCPTABLE, M_WAITOK );
  
  aux= ip6_default_srcptable;
  
  /* 1 raw */ 
  aux->port=0;
  bcopy( &entry_1, &aux->prefix, 16 );
  /*inet_pton6( "::1", &aux->prefix );*/
  aux->prefixlen= 128;
  aux->precedence=50;
  aux->label=0;
  aux->rth= in6addr_any;

  /* 2 raw */
  aux->next=(struct ip6_srcptable *)
    malloc( sizeof(struct ip6_srcptable), M_SRCPTABLE, M_WAITOK );
  aux= aux->next;
  aux->port=0;
  aux->prefix= in6addr_any;
  aux->prefixlen= 0;
  aux->precedence=40;
  aux->label=1;
  aux->rth= in6addr_any;

  /* 3 raw */
  aux->next=(struct ip6_srcptable *)
    malloc( sizeof(struct ip6_srcptable), M_SRCPTABLE, M_WAITOK );
  aux= aux->next;
  aux->port=0;
  bcopy( &entry_2, &aux->prefix, 16 );       
  /*inet_pton6( "2002::0", &aux->prefix );*/
  aux->prefixlen= 16;
  aux->precedence=30;
  aux->label=2;
  aux->rth= in6addr_any;

  /* 4 raw */
  aux->next=(struct ip6_srcptable *)
    malloc( sizeof(struct ip6_srcptable), M_SRCPTABLE, M_WAITOK );
  aux= aux->next;
  aux->port=0;
  aux->prefix= in6addr_any;
  aux->prefixlen= 96;
  aux->precedence=20;
  aux->label=3;
  aux->rth= in6addr_any;

  /* 5 raw */
  aux->next=(struct ip6_srcptable *)
    malloc( sizeof(struct ip6_srcptable), M_SRCPTABLE, M_WAITOK );
  aux= aux->next;
  aux->port=0;
  bcopy( &entry_3, &aux->prefix, 16 );       
  /*inet_pton6( "::ffff:0:0", &aux->prefix );*/
  aux->prefixlen= 96;
  aux->precedence=10;
  aux->label= 4;
  aux->rth= in6addr_any;
  ;

  /* final */
  aux->next = 0;
}


int ip6_add_srcptable( newpt )
     struct ip6_srcptable *newpt;
{
  struct ip6_srcptable *aux, *new_entry;

  new_entry= (struct ip6_srcptable *)
    malloc( sizeof(struct ip6_srcptable), M_SRCPTABLE, M_WAITOK );
  bcopy( newpt, new_entry, sizeof(struct ip6_srcptable) );
  new_entry->next = 0;

  if( ip6_default_srcptable == 0 )  
    ip6_default_srcptable = new_entry;  
  else
  {
    aux = ip6_default_srcptable;
    ip6_default_srcptable = new_entry;
    ip6_default_srcptable->next = aux;
  }

  return 0;
}

int ip6_del_srcptable( oldpt )     
     struct ip6_srcptable *oldpt;
{
  struct ip6_srcptable *aux, *ant= 0;

  aux = ip6_default_srcptable;

  while( aux != 0 )
  {
    if( ip6_srcptable_equals( aux, oldpt ) )
      break;

    ant= aux;
    aux= aux->next;
  }

  if( aux != 0 )
  {
    if( aux == ip6_default_srcptable )
      {
	ip6_default_srcptable= ip6_default_srcptable->next;
	free( aux, M_SRCPTABLE );
      }
    else
      {
	ant->next= aux->next;
	free( aux, M_SRCPTABLE );
      }
  }
  else
   return EINVAL;

  return 0;
}

int ip6_srcptable_equals( aux1, aux2 ) 
     struct ip6_srcptable *aux1, *aux2;          
{
  if( aux1->port != aux2->port )
    return 0;
  else
    if( ! IN6_ARE_ADDR_EQUAL(&aux1->prefix, &aux2->prefix) )
      return 0;
  else
    if( aux1->prefixlen != aux2->prefixlen )
      return 0;
  else
    if( aux1->precedence != aux2->precedence )
      return 0;
  else
    if( aux1->label != aux2->label )
      return 0;
  else
    if( ! IN6_ARE_ADDR_EQUAL(&aux1->rth, &aux2->rth) )
      return 0;

  return 1;
}


/* return the label and optionally, the routing header in last argument.
   TODO: if used in destination address selection algorithm, return also
   the precedence. */
int ip6_search_srcptable( port, addr, addrlen, rth )
    int port, addrlen;
    struct in6_addr *addr, *rth;
    /* rth can be NULL !! */
{
  struct ip6_srcptable *aux;
  int i, j, k, blen, tlen;
  struct ip6_srcptable *has_eq_port[256], 
    *without_port[256], *best_entry= 0; /*FIXME: hard-coded */

  /**************************************************/
  /* PASS ONE: classsify the entrys into the arrays */
  /**************************************************/
  aux= ip6_default_srcptable;
  i= 0;
  j= 0;
  while( aux != 0)
  {
    if( aux->port != 0 )
    {
      if( aux->port == port )
	has_eq_port[i++]= aux;
    }
    else
      without_port[j++]= aux;

     aux= aux->next;
  }

  /*****************************************************************************/
  /* PASS TWO: search the port entries only and select by longer prefix match. */
  /*****************************************************************************/

  /* NOTE: If there is not match, perform the last search with
     entries that has not port. */
  blen= -1;  
  for(k=0; k<i;k++)
  {  
    tlen= in6_matchlen( &has_eq_port[k]->prefix, addr );

    /* adjust the match in the same manner that we'd do if this were
       a routing table, for example, if the match was 36 bits, and the entry
       had a prefix of 32, the resulting match is 32, unless there is another
       entry with a prefixlen of 36, in which case the match will be 36. Note
       that if the match were 36 and the prefixlen entry were 38, for example, 
       there would be no match ! */

    /* restrict the match to the prefixlen of the address. */
    if( tlen > addrlen )
      tlen = addrlen;

    /* like a routing table, there is a match only when tlen >= entry_len */
    if( tlen >= has_eq_port[k]->prefixlen )
      tlen = has_eq_port[k]->prefixlen;
    else
      tlen = -1;	

    if( tlen > blen )
    {
      blen= tlen;
      best_entry= has_eq_port[k];      
    }
  }
 
  if( blen >= 0)
    goto end;

  /***********************************************************************************/
  /* PASS THREE: search the entries without ports and select by longer prefix match. */
  /***********************************************************************************/ 
  for(k=0; k<j;k++)
  {
    tlen= in6_matchlen( &without_port[k]->prefix, addr );

    /* adjust the match in the same manner that we'd do if this were
       a routing table, for example, if the match was 36 bits, and the entry
       had a prefix of 32, the resulting match is 32, unless there is another
       entry with a prefixlen of 36, in which case the match will be 36. Note
       that if the match were 36 and the prefixlen entry were 38, for example, 
       there would be no match ! */

    /* restrict the match to the prefixlen of the address. */
    if( tlen > addrlen )
      tlen = addrlen;

    /* like a routing table, there is a match only when tlen >= entry_len */
    if( tlen >= without_port[k]->prefixlen )
      tlen = without_port[k]->prefixlen;
    else
      tlen = -1;	

    if( tlen > blen )
    {
      blen= tlen;
      best_entry= without_port[k];     
    }      
  }

  /**************************************************************************/
  /* FINAL STEP: it is different to have a match of zero than to have no match ! */
  /**************************************************************************/
  if( blen == -1) 
  {   
    if( rth != NULL )
      bcopy( &in6addr_any, rth, sizeof(struct in6_addr));

    return -1; /* it means no-label, no-match */
  }

 end:
  if( rth != NULL )
    bcopy( (char *)&best_entry->rth, (char *)rth, sizeof(struct in6_addr));

  return best_entry->label;  
}



