#include <stdio.h>
#include <stdlib.h>                

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>         
#include <netinet6/qospsas.h>

#define LINE 80

#define ADD_OPTION 1
#define DEL_OPTION 2
#define SHOW_OPTION 3


char blancos[INET6_ADDRSTRLEN];

void print_ptable( struct ip6_srcptable *aux ) 
{ 
  char str_addr[INET6_ADDRSTRLEN], str_addr2[INET6_ADDRSTRLEN];
		
  inet_ntop( AF_INET6, &aux->prefix, str_addr, INET6_ADDRSTRLEN );
  sprintf( str_addr+strlen(str_addr), "/%d", aux->prefixlen );
  bcopy( blancos, str_addr+strlen(str_addr), INET6_ADDRSTRLEN - strlen(str_addr) - 1 );
  str_addr[INET6_ADDRSTRLEN-1]= '\0';

  if( IN6_IS_ADDR_UNSPECIFIED(&aux->rth) )    
  {
    if( aux->port == 0 )
      printf("%s\t%s   %d\t%d\t%s\n", "-", str_addr, 
	      aux->precedence, aux->label, "-" );  
    else
      printf("%d\t%s   %d\t%d\t%s\n", aux->port, str_addr, 
	     aux->precedence, aux->label, "-" );  
  }
  else
  {
    inet_ntop( AF_INET6, &aux->rth, str_addr2, INET6_ADDRSTRLEN );  
    if( aux->port == 0 )
      printf("%s\t%s   %d\t%d\t%s\n", "-", str_addr, 
	     aux->precedence, aux->label, str_addr2 );
    else
      printf("%d\t%s   %d\t%d\t%s\n", aux->port, str_addr, 
	     aux->precedence, aux->label, str_addr2 );
  } 
}

void print_help()
{
  printf("Usage:\n");
  printf("\tpolicy-table add|del [-port <num>] -prefix <addr/len> -preced <num> "
	                 "-label <num> [-rth <addr>]\n");
  printf("\tpolicy-table show\n\n");
  exit(-1);
}

void parse_suboptions( int argc, char *argv[], struct ip6_srcptable *ptable )
{ 
  int optind= 0;

  /* store values for optional data  in the buffer. */
  bcopy( in6addr_any, &ptable->rth, sizeof(struct in6_addr) );  
  ptable->port= 0;
  
  while( optind < argc )
  { 
    if( ! strcasecmp("-port", argv[optind] ) )
    {
      ptable->port= atoi( argv[++optind] ); 
      if( ptable->port == 0 )
      {
	printf("error: port can not be zero.");
	exit(-1);
      }
    }
    else
    if( ! strcasecmp("-prefix", argv[optind] ) )
    {  
      if( inet_pton( AF_INET6, (char *)strtok( argv[++optind], "/"), &ptable->prefix ) < 0)
      {
	perror("inet_pton error:");
	exit(-1);
      }

      ptable->prefixlen= atoi( (char *)strtok(NULL, "/") );     
      if( (ptable->prefixlen < 0) || (ptable->prefixlen > 128) )
      {
	perror("error: the prefix length mus be between [0,128]");
	exit(-1);
      }
    }
    else
    if( ! strcasecmp("-preced", argv[optind] ) )          
    {
      ptable->precedence = atoi( argv[++optind] );  
    }
    else
    if( ! strcasecmp("-label", argv[optind] ) )
    {  
      ptable->label = atoi( argv[++optind] );
    }
    else
    if( ! strcasecmp("-rth", argv[optind] ) )
    {      
      if( inet_pton( AF_INET6, argv[++optind], &ptable->rth ) < 0)
      {
	perror("inet_pton error:");
	exit(-1);
      }
      if( IN6_IS_ADDR_UNSPECIFIED( &ptable->rth ) )
      {
	printf("error: routing header can not be unspecified\n");
	exit(-1);
      }
    }
    else   
      print_help();
    
    optind++;
  }
}

int main(int argc,  char *argv[])
{
  struct ip6_srcptable *ptable, *bb, *be;
  char op[LINE], op2[LINE];
  char aux[INET6_ADDRSTRLEN], buffer[1024]; /* XXX: hard-coded */
  int buf_len, i, c; 
  int option= 0;

  /* initialize some buffers & pointers.*/
  for(i=0;i<INET6_ADDRSTRLEN;i++)
    blancos[i]=' ';

  ptable = (struct ip6_srcptable *)(buffer+1);

  /* parse the first argument, only can be "add", "del" or "show" */
  if( argc < 2 )
    print_help();

  if( ! strcasecmp("add", argv[1] ) )
  {
    if( (argc != 8) && (argc != 10) && (argc != 12) )
      print_help();
    option= ADD_OPTION;
  }
  else 
  if( ! strcasecmp("del", argv[1] ) )
  {
    if( (argc != 8) && (argc != 10) && (argc != 12) )
      print_help();
    option= DEL_OPTION;
  }
  else
  if( ! strcasecmp("show", argv[1] ) )
  {
    if( argc != 2 )
      print_help();
    option= SHOW_OPTION;
  }
  else
    print_help();

  /* if "add/del", parse remaining arguments */
  switch (option)
  {
  case ADD_OPTION:
  case DEL_OPTION:
    parse_suboptions( argc - 2, argv + 2, ptable );
    buf_len= sizeof(struct ip6_srcptable);  
    if( option == ADD_OPTION )
    {
      buffer[0]='A';
      if( sysctlbyname( "net.inet6.ip6.srcptable", NULL, 0, buffer, &buf_len  ) )
      {
	perror("Error al aadir: ");
	exit(-1);
      }    
      printf("policy-table: policy-table entry added succesfully.\n\n");
    }
    else
    {
      buffer[0]='D';
      if( sysctlbyname( "net.inet6.ip6.srcptable", NULL, 0, buffer, &buf_len  ) )
      {
	perror("Error al borrar: ");
	exit(-1);
      }    
      printf("policy-table: policy-table entry deleted succesfully.\n\n");
    }
    break;
  case SHOW_OPTION: 
    buf_len= 1024;
    if( sysctlbyname( "net.inet6.ip6.srcptable", buffer, &buf_len, NULL, 0 ) )
    {
      perror("policy-table error: ");
      exit(-1);
    }
    printf("\npolicy_table:\n");
    printf("port\tprefix\t\t\t\t\t\tpreced\tlabel\tRH\n");  
    
    be= (struct ip6_srcptable *)(buffer+buf_len);

    for( bb = (struct ip6_srcptable *)buffer; bb + 1 <= be; bb = bb + 1 )     
      print_ptable( bb );      

    putchar('\n');
    break;
  }
    
  return 0;
}
